
// fx - A compositing library representation.

// Copyright (C) 2005-2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifndef RENDER_BUFFER_INC_
#define RENDER_BUFFER_INC_

#include <openeffectslib/fx/config.hpp>

#include <string>

#include <boost/shared_ptr.hpp>

#include <GL/glew.h>

#ifdef WIN32
#include <GL/wglew.h>
#else
#include <GL/glxew.h>
#endif

#include <openimagelib/il/il.hpp>

#include <openeffectslib/fx/ogl_utilities.hpp>

namespace oil = olib::openimagelib::il;

namespace olib { namespace openeffectslib { namespace fx {

//
class FX_DECLSPEC default_surface_format
{
public:
	explicit default_surface_format( const std::string& options );
	
public:
	int r_bits( ) const { return r_bits_; }
	int g_bits( ) const { return g_bits_; }
	int b_bits( ) const { return b_bits_; }
	int a_bits( ) const { return a_bits_; }	

public:
	int depth_bits( )   const { return depth_bits_; }
	int stencil_bits( ) const { return stencil_bits_; }
	
public:
	bool is_float( )	 const { return float_; }
	bool doublebuffer( ) const { return doublebuffer_; }
	bool mipmap( )		 const { return mipmap_; }
	
public:
	std::string target( ) const { return target_; }

private:
	int  r_bits_, g_bits_, b_bits_, a_bits_;
	int  depth_bits_, stencil_bits_;
	bool float_, doublebuffer_, mipmap_;
	std::string target_;
};

// mix-in class for different buffer types.
template<class surface_format>
#if _MSC_VER < 1400 // some dodgy stuff going on.
class buffer_type : public default_surface_format
#else
class buffer_type : public surface_format
#endif
{
public:
	explicit buffer_type( const std::string& options = "" )
#if _MSC_VER < 1400
		: default_surface_format( options )
#else
		: surface_format( options )
#endif
	{ }

	virtual ~buffer_type( )
	{ }

public:
	virtual bool				init( int /*width*/, int /*height*/, int /*pw*/, int /*ph*/, bool /*share*/ )	{ return false; }
	virtual bool				init( const std::wstring /*colour_space*/, int /*height*/, int /*pw*/, int /*ph*/, bool /*share*/ )	{ return false; }
	virtual bool				init( const oil::image_type_ptr /*im*/, int /*pw*/, int /*ph*/ ) { return false; }
	virtual void				bind_as_GL_texture( ) const					= 0;
	virtual void				unbind_as_GL_texture( ) const				= 0;
	virtual GLuint				get_GL_texture( ) const						= 0;
	virtual oil::image_type_ptr	data( )	const								= 0;
	virtual bool				initialised( ) const						= 0;
	
protected:
	virtual bool begin_render_to( bool clear ) const			= 0;
	virtual bool end_render_to( ) const							= 0;

public:
	bool power_of_two( ) const
	{ return ( !( width_ & ( width_ - 1 ) ) && !( height_ & ( height_ - 1 ) ) ); }

	GLenum GL_target( ) const
	{
#if _MSC_VER < 1400
		if( default_surface_format::target( ) == "tex2D" || default_surface_format::target( ) == "texRECT" )
#else
		if( surface_format::target( ) == "tex2D" || default_surface_format::target( ) == "texRECT" )
#endif
		{
			// we should really have the test for the GL version but newer ATI drivers report 2.0 and we're having problems with it.
			// their behaviour is definitely different from that of NVIDIA. Lack of GL_ARB_texture_non_power_of_two equivalent support? - Goncalo
			// I was obviously wrong. The texture coordinates were not being set correctly. - Goncalo
			if( GLEW_VERSION_2_0 || GLEW_ARB_texture_non_power_of_two || power_of_two( ) ) return GL_TEXTURE_2D;
			else if( GLEW_ARB_texture_rectangle )	return GL_TEXTURE_RECTANGLE_ARB;
			else if( GLEW_EXT_texture_rectangle )	return GL_TEXTURE_RECTANGLE_EXT;
			else if( GLEW_NV_texture_rectangle )	return GL_TEXTURE_RECTANGLE_NV;
			else assert( 0 && L"texturing not available?!" );
		}
#if _MSC_VER < 1400
		else if( default_surface_format::target( ) == "texCUBE" )
#else
		else if( surface_format::target( ) == "texCUBE" )
#endif
		{
		}
		
		return 0;
	}
	
public:
	int width( )  const { return width_;  }
	int height( ) const { return height_; }

protected:
	int width_, height_;
};

//
template<class surface_format>
class pbuffer : public buffer_type<surface_format>
{
public:
	explicit pbuffer( const std::string& options = "" )
#ifdef WIN32
		: hdc_( NULL )
		, old_hdc_( NULL )
		, hglrc_( NULL )
		, old_hglrc_( NULL )
		, pbuffer_( NULL )
#endif
	{ }
	
	virtual ~pbuffer( )
	{ }

private:
#ifdef WIN32
	HDC			hdc_;
	HDC			old_hdc_;
	HGLRC		hglrc_;
	HGLRC		old_hglrc_;
	HPBUFFERARB pbuffer_;
#else
#endif
};

//
template<class surface_format = default_surface_format>
class framebuffer_object : public buffer_type<surface_format>
{
public:
	explicit framebuffer_object( const std::string& options = "" )
		: buffer_type<surface_format>( options )
		, fb_id_( 0 )
		, rb_id_( 0 )
		, tex_id_( 0 )
		, clear_flags_( GL_COLOR_BUFFER_BIT )
		, p_max_s_( 0.0f )
		, p_max_t_( 0.0f )
		, max_s_( 0.0f )
		, max_t_( 0.0f )
		, initialised_( false )
	{ }

	virtual ~framebuffer_object( )
	{ }

public:
	virtual bool init( int width, int height, int pw = -1, int ph = -1, bool /*share*/ = true )
	{
		if( !GLEW_EXT_framebuffer_object ) return false;

		if( initialised( ) ) reset( );

		clear_flags_ = GL_COLOR_BUFFER_BIT;

		buffer_type<surface_format>::width_  = width;
		buffer_type<surface_format>::height_ = height;

		glGenFramebuffersEXT( 1, &fb_id_ );
		glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fb_id_ );

		glGenTextures( 1, &tex_id_ );
		glBindTexture( buffer_type<surface_format>::GL_target( ), tex_id_ );
		glTexParameteri( buffer_type<surface_format>::GL_target( ), GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( buffer_type<surface_format>::GL_target( ), GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glTexParameteri( buffer_type<surface_format>::GL_target( ), GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( buffer_type<surface_format>::GL_target( ), GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
		glTexImage2D( buffer_type<surface_format>::GL_target( ), 0, GL_RGB, width, height, 0, GL_RGB, buffer_type<surface_format>::is_float( ) ? GL_FLOAT : GL_UNSIGNED_BYTE, 0 );

		if( buffer_type<surface_format>::mipmap( ) ) glGenerateMipmapEXT( buffer_type<surface_format>::GL_target( ) );
		glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, buffer_type<surface_format>::GL_target( ), tex_id_, 0 );

		if( buffer_type<surface_format>::doublebuffer( ) )
		{
			glGenRenderbuffersEXT( 1, &rb_id_ );
			glBindRenderbufferEXT( GL_RENDERBUFFER_EXT, rb_id_ );
			glRenderbufferStorageEXT( GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, width, height );
			glFramebufferRenderbufferEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, rb_id_ );

			clear_flags_ |= GL_DEPTH_BUFFER_BIT;
		}

		GLenum status;
		if( ( status = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT ) ) != GL_FRAMEBUFFER_COMPLETE_EXT )
		{
			reset( );
		
			return false;
		}

		p_max_s_ = buffer_type<surface_format>::GL_target( ) == GL_TEXTURE_2D ? 1.0f : ( pw != -1 ? pw : width );
		p_max_t_ = buffer_type<surface_format>::GL_target( ) == GL_TEXTURE_2D ? 1.0f : ( ph != -1 ? ph : height );
		max_s_ = buffer_type<surface_format>::GL_target( ) == GL_TEXTURE_2D ? 1.0f : width;
		max_t_ = buffer_type<surface_format>::GL_target( ) == GL_TEXTURE_2D ? 1.0f : height;

		glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

		initialised_ = true;

		return true;
	}

private:
	void reset( )
	{
		glDeleteTextures( 1, &tex_id_ );
		glDeleteRenderbuffersEXT( 1, &rb_id_ );
		glDeleteFramebuffersEXT( 1, &fb_id_ );
		
		tex_id_ = 0;
		rb_id_  = 0;
		fb_id_  = 0;
		
		initialised_ = false;
	}

public:
	virtual bool begin_render_to( bool clear ) const
	{
		glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fb_id_ );
		glViewport( 0, 0, buffer_type<surface_format>::width( ), buffer_type<surface_format>::height( ) );
		
		if( clear ) glClear( clear_flags_ );
	
		glMatrixMode( GL_PROJECTION );
		glPushMatrix( );
		glLoadIdentity( );
		gluOrtho2D( 0.0f, buffer_type<surface_format>::width( ), 0.0f, buffer_type<surface_format>::height( ) );
		
		glMatrixMode( GL_MODELVIEW );
		glPushMatrix( );
		glLoadIdentity( );

		return true;
	}

	virtual bool end_render_to( ) const
	{
		glMatrixMode( GL_PROJECTION );
		glPopMatrix( );
	
		glMatrixMode( GL_MODELVIEW );
		glPopMatrix( );

		glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
		
		return true;
	}

public:
	virtual void bind_as_GL_texture( ) const
	{
		glEnable( buffer_type<surface_format>::GL_target( ) );
		glBindTexture( buffer_type<surface_format>::GL_target( ), tex_id_ );
	}

	virtual void unbind_as_GL_texture( ) const
	{
		glBindTexture( buffer_type<surface_format>::GL_target( ), 0 );
		glDisable( buffer_type<surface_format>::GL_target( ) );
	}
	
	virtual GLuint get_GL_texture( ) const
	{ return tex_id_; }

	virtual GLenum get_GL_target( ) const
	{ return buffer_type<surface_format>::GL_target( ); }
	
	virtual oil::image_type_ptr data( ) const { return oil::image_type_ptr( ); }
	
	float physical_max_s( ) const	{ return p_max_s_; }
	float physical_max_t( ) const	{ return p_max_t_; }
	float max_s( ) const			{ return max_s_; }
	float max_t( ) const			{ return max_t_; }
	
public:
	virtual bool initialised( ) const
	{ return initialised_; }
	
private:
	GLuint fb_id_, rb_id_, tex_id_;
	GLenum clear_flags_;
	float p_max_s_, p_max_t_;
	float max_s_, max_t_;
	bool initialised_;
};

//
template<class surface_format>
class memory : public buffer_type<surface_format>
{
public:
	explicit memory( const std::string& options = "" )
		: buffer_type<surface_format>( options )
		, id_( 0 )
		, p_max_s_( 0.0f )
		, p_max_t_( 0.0f )
		, max_s_( 0.0f )
		, max_t_( 0.0f )
		, old_phy_w_( 0 )
		, old_phy_h_( 0 )
		, initialised_( false )
	{ }

	virtual ~memory( )
	{ }

public:
	virtual bool init( const oil::image_type_ptr im, int pw = -1, int ph = -1 )
	{
		if( !im ) return false;

		if( !glIsTexture( id_ ) )
			glGenTextures( 1, &id_ );

		if( data_ )
		{
			old_phy_w_ = data_->width( );
			old_phy_h_ = data_->height( );
		}
		
		int width  = im->width( );
		int height = im->height( );
		
		if( width != old_phy_w_ || height != old_phy_h_ )
		{
			buffer_type<surface_format>::width_  = width;
			buffer_type<surface_format>::height_ = height;
		
			GLenum target = buffer_type<surface_format>::GL_target( );
		
			p_max_s_ = target == GL_TEXTURE_2D ? 1.0f : ( pw != -1 ? pw : width );
			p_max_t_ = target == GL_TEXTURE_2D ? 1.0f : ( ph != -1 ? ph : height );
			max_s_	 = target == GL_TEXTURE_2D ? 1.0f : width;
			max_t_	 = target == GL_TEXTURE_2D ? 1.0f : height;
		}
		
		data_ = im;
		initialised_ = true;

		return true;
	}

public:
	virtual bool begin_render_to( bool clear ) const	{ return true; }
	virtual bool end_render_to( bool clear ) const		{ return true; }

	virtual void bind_as_GL_texture( ) const
	{
		GLenum target = buffer_type<surface_format>::GL_target( );

		typedef std::pair<GLenum, GLenum> internal_and_base_format;
		internal_and_base_format pf = fx::pf_to_gl_format( data_->pf( ) );

		glEnable( target );
		glBindTexture( target, id_ );
		
		glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );
		glTexParameteri( target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		int phy_w = data_->width( );
		int phy_h = data_->height( );

		if( phy_w != old_phy_w_ || phy_h != old_phy_h_ )
		{
			glTexParameteri( target, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
			glTexParameteri( target, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
			glTexImage2D( target, 0, pf.first, phy_w, phy_h, 0, pf.second, GL_UNSIGNED_BYTE, data_->data( ) );
		}
		else
		{
			glTexSubImage2D( target, 0, 0, 0, phy_w, phy_h, pf.second, GL_UNSIGNED_BYTE, data_->data( ) );
		}
	
		glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	}

	virtual void unbind_as_GL_texture( ) const
	{
		GLenum target = buffer_type<surface_format>::GL_target( );
		
		glBindTexture( target, 0 );
		glDisable( target );
	}
	
	virtual GLuint get_GL_texture( ) const
	{ return id_; }

	virtual GLenum get_GL_target( ) const
	{ return buffer_type<surface_format>::GL_target( ); }
	
	virtual oil::image_type_ptr data( ) const { return data_; }
	
	float physical_max_s( ) const	{ return p_max_s_; }
	float physical_max_t( ) const	{ return p_max_t_; }
	float max_s( ) const			{ return max_s_; }
	float max_t( ) const			{ return max_t_; }

public:
	virtual bool initialised( ) const
	{ return initialised_; }
	
private:
	oil::image_type_ptr data_;
	GLuint id_;
	float p_max_s_, p_max_t_;
	float max_s_, max_t_;
	int old_phy_w_, old_phy_h_;
	bool initialised_;
};

//
template<template<class> class buffer_type, class surface_format>
class render_buffer
{
public:
	explicit render_buffer( const std::string& options = "" )
		: structure_( new buffer_type<surface_format>( options ) )
	{ }

	void swap( render_buffer& other )
	{ structure_.swap( other.structure_ ); }

public:
	bool init( int w, int h, int pw = -1, int ph = -1, bool share = true )	{ return structure_->init( w, h, pw, ph, share ); }
	bool begin_render_to( bool clear = true ) const							{ return structure_->begin_render_to( clear ); }
	bool end_render_to( ) const												{ return structure_->end_render_to( ); }
	void bind_as_GL_texture( ) const										{ return structure_->bind_as_GL_texture( ); }
	void unbind_as_GL_texture( ) const										{ return structure_->unbind_as_GL_texture( ); }
	GLuint get_GL_texture( ) const											{ return structure_->get_GL_texture( ); }
	GLenum get_GL_target( ) const											{ return structure_->get_GL_target( ); }
	bool initialised( ) const												{ return structure_->initialised( ); }

public:
	int width( )  const { return structure_->width( );  }
	int height( ) const { return structure_->height( ); }

	float physical_max_s( ) const { return structure_->physical_max_s( ); }
	float physical_max_t( ) const { return structure_->physical_max_t( ); }
	float max_s( ) const { return structure_->max_s( ); }
	float max_t( ) const { return structure_->max_t( ); }
	
private:
	boost::shared_ptr<buffer_type<surface_format> > structure_;
};

} } }

#endif
